from typing import Tuple

import cv2
import numpy as np
from typing import List


from ..visualization.color import Color


class Drawing:
    @staticmethod
    def circle(
        image: np.ndarray,
        center_coordinates: Tuple[int, int],
        thickness: int = -1,
        color: Tuple[int, int, int] = Color.black,
        radius: int = 3,
    ) -> np.ndarray:
        """
        Draws a circle on an image.

        Args:
            image (np.ndarray): The image on which to draw the circle.
            center_coordinates (Tuple[int, int]): The center coordinates of the circle.
            thickness (int): Thickness of the circle outline. Default is -1 (filled circle).
            color (Tuple[int, int, int]): Color of the circle in BGR format.
            radius (int): Radius of the circle.

        Returns:
            np.ndarray: The image with the drawn circle.
        """
        # Ensure the image has the correct data type
        if image.dtype != np.uint8:
            image = image.astype(np.uint8)

        image = cv2.circle(image, center_coordinates, radius, color, thickness)

        return image

    @staticmethod
    def put_text(
        text: str,
        image: np.ndarray,
        org: Tuple[int, int],
        color: Tuple[int, int, int] = (0, 0, 0),
        fontScale: float = 0.25,
    ) -> np.ndarray:
        """
        Places text on an image.

        Args:
            text (str): The text string to be drawn.
            image (np.ndarray): The image on which to draw the text.
            org (Tuple[int, int]): The bottom-left corner of the text string in the image.
            color (Tuple[int, int, int]): Color of the text in BGR format.
            fontScale (float): Font scale factor.

        Returns:
            np.ndarray: The image with the drawn text.
        """
        font = cv2.FONT_HERSHEY_DUPLEX
        thickness = 1
        image = cv2.putText(
            image, text, org, font, fontScale, color, thickness, cv2.LINE_AA
        )

        return image

    @staticmethod
    def intersection(
        dot: "Node",
        img: np.ndarray,
        color: Tuple[int, int, int] = Color.red,
    ) -> np.ndarray:
        """
        Marks an intersection point on an image.

        Args:
            dot (Node): The point to mark, expected to have attributes 'x' and 'y'.
            img (np.ndarray): The image on which to mark the point.
            color (Tuple[int, int, int]): Color of the mark in BGR format.

        Returns:
            np.ndarray: The image with the marked intersection.
        """
        from ..geometry.node import Node

        img[int(dot.y), int(dot.x), :] = color

        return img

    @staticmethod
    def curve(
        curve: "Curve",
        img: np.ndarray,
        color: Tuple[int, int, int] = (0, 255, 0),
        thickness: int = 1,
    ) -> np.ndarray:
        """
        Draws a curve on an image.

        Args:
            curve (Curve): The curve object, expected to have 'geometry.xy' attributes.
            img (np.ndarray): The image on which to draw the curve.
            color (Tuple[int, int, int]): Color of the curve in BGR format.
            thickness (int): Thickness of the curve lines.

        Returns:
            np.ndarray: The image with the drawn curve.
        """
        from ..geometry.curve import Curve

        y, x = curve.geometry.xy
        y = np.array(y).astype(int)
        x = np.array(x).astype(int)
        pts = np.vstack((x, y)).T
        isClosed = False
        img = cv2.polylines(img, [pts], isClosed, color, thickness)

        return img

    @staticmethod
    def chain(
        chain: "Chain",
        img: np.ndarray,
        color: Tuple[int, int, int] = (0, 255, 0),
        thickness: int = 5,
    ) -> np.ndarray:
        """
        Draws a chain of nodes on an image.

        Args:
            chain (Chain): The chain object, expected to have 'get_nodes_coordinates()' method.
            img (np.ndarray): The image on which to draw the chain.
            color (Tuple[int, int, int]): Color of the chain in BGR format.
            thickness (int): Thickness of the chain lines.

        Returns:
            np.ndarray: The image with the drawn chain.
        """
        from ..geometry.chain import Chain

        y, x = chain.get_nodes_coordinates()
        pts = np.vstack((y, x)).T.astype(int)
        isClosed = False
        img = cv2.polylines(img, [pts], isClosed, color, thickness)

        return img

    @staticmethod
    def radii(
        ray: "Ray",
        img: np.ndarray,
        color: Tuple[int, int, int] = (255, 0, 0),
        thickness: int = 1,
    ) -> np.ndarray:
        """
        Draws a radius line on an image.

        Args:
            ray (Radii): The radii object, expected to have 'geometry.xy' attributes.
            img (np.ndarray): The image on which to draw the radius.
            color (Tuple[int, int, int]): Color of the radius line in BGR format.
            thickness (int): Thickness of the radius line.

        Returns:
            np.ndarray: The image with the drawn radius.
        """
        from ..geometry.ray import Ray

        y, x = ray.geometry.xy
        y = np.array(y).astype(int)
        x = np.array(x).astype(int)
        start_point = (x[0], y[0])
        end_point = (x[1], y[1])
        image = cv2.line(img, start_point, end_point, color, thickness)

        return image

    @staticmethod
    def write_curves_to_image(
        curves_list: List["Curve"], img: np.ndarray
    ) -> np.ndarray:
        """
        Draws curves onto an image.

        Args:
            curves_list (List[Curve]): Array of curve points.
            img (np.ndarray): Image array.

        Returns:
            np.ndarray: Image with curves drawn.
        """
        from ..geometry.curve import Curve

        img_aux = np.full((img.shape[0], img.shape[1]), 255)

        for pix in curves_list:
            if pix[0] < 0 and pix[1] < 0:
                continue
            img_aux = Drawing.circle(img_aux, (int(pix[0]), int(pix[1])))

        return img_aux



from dataclasses import dataclass, field
from pathlib import Path
import json
import logging
from typing import Any, Dict, Optional
from datetime import datetime

from .utils.file_utils import ensure_directory

logger = logging.getLogger(__name__)


@dataclass
class Config:
    """Global configuration settings for tree ring detection.

    Tracks changes to settings and provides validation for paths.
    All changes to settings are automatically logged.
    """

    # -------------- Input/Output Settings ----------------

    input_image: Optional[Path] = None
    """Path to the input image file."""

    output_dir: Path = Path("./output/")
    """Directory where results and debug information will be saved."""

    # -------------- Image Processing Parameters ----------------

    cx: Optional[int] = None
    """Center x-coordinate in the image."""

    cy: Optional[int] = None
    """Center y-coordinate in the image."""

    sigma: float = 3.0
    """Gaussian kernel parameter for Canny edge detector. Controls edge smoothing."""

    output_height: Optional[int] = None
    """Target height for image resizing. If 0, maintains original height."""

    output_width: Optional[int] = None
    """Target width for image resizing. If 0, maintains original width."""

    # -------------- Edge Detection Parameters ----------------

    th_low: float = 5.0
    """Low threshold on gradient magnitude for Canny edge detector. Controls edge sensitivity."""

    th_high: float = 20.0
    """High threshold on gradient magnitude for Canny edge detector. Controls edge continuity."""

    alpha: float = 30.0
    """Edge filtering parameter (collinearity threshold) in degrees."""

    # -------------- Processing Settings ----------------
    nr: int = 360
    """Number of rays for sampling. Higher values give more precise detection."""

    min_chain_length: int = 2
    """Minimum chain length. Chains shorter than this are filtered out."""

    # -------------- Operation Modes ----------------
    debug: bool = False
    """Enable debug mode for additional logging and visualizations."""

    save_results: bool = False
    """Save intermediate images, labelme and config file."""

    clear_output: bool = True
    """Whether to clear output directory before use."""

    _change_history: Dict[str, list] = field(default_factory=dict, repr=False)
    """Change history tracking."""

    def __post_init__(self):
        """Initialize paths and change history tracking."""
        self._validate_and_set_paths()
        self._change_history = {}
        for field_name in self.__dataclass_fields__:
            if not field_name.startswith("_"):
                self._change_history[field_name] = []

    def _validate_and_set_paths(self):
        """Validate and set all path-related fields."""
        # Set up output directory
        output_path = Path(self.output_dir)
        try:
            self.output_dir = ensure_directory(output_path, clear=self.clear_output)
        except PermissionError:
            raise ValueError(
                f"Cannot create/clear output directory (permission denied): {output_path}"
            )
        except Exception as e:
            raise ValueError(f"Error with output directory: {output_path}, {str(e)}")

        # Validate input image if provided
        if self.input_image:
            input_path = Path(self.input_image)

            if not input_path.exists():
                raise ValueError(f"Input image file does not exist: {input_path}")
            if not input_path.is_file():
                raise ValueError(f"Input image path is not a file: {input_path}")

            self.input_image = input_path.resolve()

    def _log_change(self, param: str, old_value: Any, new_value: Any):
        """Log a parameter change with timestamp."""
        timestamp = datetime.now().isoformat()
        change_record = {
            "timestamp": timestamp,
            "old_value": old_value,
            "new_value": new_value,
        }
        self._change_history[param].append(change_record)
        logger.info(f"Config change: {param} changed from {old_value} to {new_value}")

    def update(self, **kwargs):
        """
        Update configuration with new values and log changes.

        Args:
            **kwargs: Configuration parameters to update.

        Raises:
            ValueError: If parameter doesn't exist or paths are invalid.
        """
        path_params = {"input_image", "root_dir", "output_dir"}
        needs_validation = any(param in path_params for param in kwargs)

        for key, new_value in kwargs.items():
            if not hasattr(self, key):
                raise ValueError(f"Unknown configuration parameter: {key}")

            old_value = getattr(self, key)
            if old_value != new_value:
                setattr(self, key, new_value)
                self._log_change(key, old_value, new_value)

        if needs_validation:
            self._validate_and_set_paths()

    def get_change_history(self, param: Optional[str] = None) -> Dict:
        """
        Get change history for a specific parameter or all parameters.

        Args:
            param: Optional parameter name. If None, returns all change history.

        Returns:
            Dictionary containing change history.
        """
        if param:
            if param not in self._change_history:
                raise ValueError(f"Unknown parameter: {param}")
            return {param: self._change_history[param]}
        return self._change_history

    def to_dict(self) -> dict:
        """Convert configuration to dictionary, excluding internal fields."""
        return {
            k: str(v) if isinstance(v, Path) else v
            for k, v in self.__dict__.items()
            if not k.startswith("_")
        }

    def to_json(self) -> str:
        """Convert configuration to JSON string."""
        return json.dumps(self.to_dict(), indent=4)

    def log_all_configs(self):
        """Log all current configuration values."""
        logger.info("Current configuration values:")
        for key, value in self.__dict__.items():
            if not key.startswith("_"):  # Skip internal attributes
                logger.info(f"{key}: {value}")


# Global configuration instance
config = Config()


def configure(**kwargs):
    """
    Configure global settings for tree ring detection.

    Args:
        **kwargs: Configuration parameters to update.

    Example:
        >>> configure(
        ...     input_image="sample.jpg",
        ...     cx=100,
        ...     cy=100
        ... )
    """
    config.update(**kwargs)



import cv2
import numpy as np
from shapely.geometry import Polygon
from ..visualization.drawing import Drawing
from ..geometry.chain import Chain


class Ring:
    def __init__(self, chain: Chain, id: int) -> None:
        """
        Initializes a Ring object.

        Args:
            chain (Chain): Chain object.
            id (int): Identifier of the ring.

        Returns:
            None
        """
        # Create a list of points (coordinates) from the sorted dots in the chain
        lista_pts = [[node.x, node.y] for node in chain.sort_dots()]
        self.id = id

        # Store the Polygon as an attribute
        self.geometry = Polygon(lista_pts)

    def draw(self, image: np.ndarray) -> np.ndarray:
        """
        Draws the ring on an image.

        Args:
            image (np.ndarray): Image on which to draw.

        Returns:
            np.ndarray: Image with the ring drawn.
        """
        x, y = self.geometry.exterior.coords.xy
        lista_pts = [[i, j] for i, j in zip(y, x)]
        pts = np.ndarray(lista_pts, np.int32)

        pts = pts.reshape((-1, 1, 2))
        is_closed = True

        # Blue color in BGR
        color = (255, 0, 0)

        # thickness of 2 px
        thickness = 1

        # Using cv2.polylines() method
        # Draw a Blue polygon with
        # thickness of 1 px
        image = cv2.polylines(image, [pts], is_closed, color, thickness)

        image = Drawing.put_text(f"{self.id}", image, (int(y[0]), int(x[0])))

        return image



import argparse
from .config import Config
from . import run, configure


def parse_arguments() -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        description="Tree Ring Detection CLI",
        formatter_class=argparse.ArgumentDefaultsHelpFormatter,
    )

    # Required arguments
    required = parser.add_argument_group("required arguments")
    required.add_argument(
        "--input_image", required=True, help="Path to input image", metavar="PATH"
    )
    required.add_argument(
        "--cx", type=int, required=True, help="Pith x-coordinate", metavar="X"
    )
    required.add_argument(
        "--cy", type=int, required=True, help="Pith y-coordinate", metavar="Y"
    )

    # Optional arguments
    parser.add_argument(
        "--output_dir", default="./output", help="Output directory", metavar="DIR"
    )
    parser.add_argument(
        "--sigma",
        type=float,
        default=Config.sigma,
        help="Edge detector Gaussian kernel parameter",
        metavar="FLOAT",
    )
    parser.add_argument(
        "--th_low",
        type=float,
        default=Config.th_low,
        help="Low threshold for gradient magnitude",
        metavar="FLOAT",
    )
    parser.add_argument(
        "--th_high",
        type=float,
        default=Config.th_high,
        help="High threshold for gradient magnitude",
        metavar="FLOAT",
    )
    parser.add_argument(
        "--height",
        type=int,
        default=Config.output_height,
        help="Resize height (0 to keep original)",
        metavar="INT",
    )
    parser.add_argument(
        "--width",
        type=int,
        default=Config.output_width,
        help="Resize width (0 to keep original)",
        metavar="INT",
    )
    parser.add_argument(
        "--alpha",
        type=float,
        default=Config.alpha,
        help="Collinearity threshold in degrees",
        metavar="FLOAT",
    )
    parser.add_argument(
        "--nr",
        type=int,
        default=Config.nr,
        help="Number of rays for sampling",
        metavar="INT",
    )
    parser.add_argument(
        "--min_chain_length",
        type=int,
        default=Config.min_chain_length,
        help="Minimum chain length for filtering",
        metavar="INT",
    )
    parser.add_argument(
        "--debug", action="store_true", help="Enable debug mode with additional logging"
    )
    parser.add_argument(
        "--save_results",
        action="store_true",
        help="Save intermediate processing images",
    )

    return parser.parse_args()


def main():
    """Main CLI entry point."""
    args = parse_arguments()

    # Configure settings from CLI arguments
    configure(
        input_image=args.input_image,
        output_dir=args.output_dir,
        cx=args.cx,
        cy=args.cy,
        sigma=args.sigma,
        th_low=args.th_low,
        th_high=args.th_high,
        output_height=args.height,
        output_width=args.width,
        alpha=args.alpha,
        nr=args.nr,
        min_chain_length=args.min_chain_length,
        debug=args.debug,
        save_results=args.save_results,
    )

    run()


if __name__ == "__main__":
    main()



from typing import Any, List, Tuple

import numpy as np
from shapely.geometry import LineString


class Ray:
    def __init__(self, direction: float, center: Tuple[float, float], M: int, N: int):
        """
        Initializes a Ray object.

        Args:
            direction (float): Angle direction of the ray in degrees.
            center (Tuple[float, float]): Center coordinates (y, x).
            M (int): Height of the image.
            N (int): Width of the image.

        Returns:
            None
        """
        self.direction = direction
        self.center = center
        self.M = M
        self.N = N
        self.border = self._image_border_radii_intersection(direction, center, M, N)
        self.geometry = LineString([center, self.border])

    @staticmethod
    def _image_border_radii_intersection(
        theta: float, origin: Tuple[float, float], M: int, N: int
    ) -> Tuple[float, float]:
        """
        Computes the intersection point of a ray with the image border.

        Args:
            theta (float): Angle of the ray in degrees.
            origin (Tuple[float, float]): Origin point (y, x).
            M (int): Height of the image.
            N (int): Width of the image.

        Returns:
            Tuple[float, float]: Intersection point coordinates (y, x).
        """
        degree_to_radians = np.pi / 180
        theta = theta % 360
        yc, xc = origin
        if 0 <= theta < 45:
            ye = M - 1
            xe = np.tan(theta * degree_to_radians) * (M - 1 - yc) + xc
        elif 45 <= theta < 90:
            xe = N - 1
            ye = np.tan((90 - theta) * degree_to_radians) * (N - 1 - xc) + yc
        elif 90 <= theta < 135:
            xe = N - 1
            ye = yc - np.tan((theta - 90) * degree_to_radians) * (xe - xc)
        elif 135 <= theta < 180:
            ye = 0
            xe = np.tan((180 - theta) * degree_to_radians) * yc + xc
        elif 180 <= theta < 225:
            ye = 0
            xe = xc - np.tan((theta - 180) * degree_to_radians) * yc
        elif 225 <= theta < 270:
            xe = 0
            ye = yc - np.tan((270 - theta) * degree_to_radians) * xc
        elif 270 <= theta < 315:
            xe = 0
            ye = np.tan((theta - 270) * degree_to_radians) * xc + yc
        elif 315 <= theta < 360:
            ye = M - 1
            xe = xc - np.tan((360 - theta) * degree_to_radians) * (ye - yc)
        else:
            raise ValueError("Invalid theta value")
        return (ye, xe)



from typing import List

from ..geometry.chain import Chain
from ..geometry.node import Node


class ConnectParameters:
    """
    Class for grouping all the parameter from table 1 in the paper
    """

    iterations = 9
    params = {
        "th_radial_tolerance": [0.1, 0.2, 0.1, 0.2, 0.1, 0.2, 0.1, 0.2, 0.2],
        "neighbourhood_size": [10, 10, 22, 22, 45, 45, 22, 45, 45],
        "th_regular_derivative": [1.5, 1.5, 1.5, 1.5, 1.5, 1.5, 2, 2, 2],
        "th_distribution_size": [2, 2, 3, 3, 3, 3, 2, 3, 3],
        "derivative_from_center": [
            False,
            False,
            False,
            False,
            False,
            False,
            True,
            True,
            True,
        ],
    }

    def __init__(self, ch_s: List[Chain], nodes_s: List[Node]) -> None:
        """
        Initialize the class with the border chain and the chains without the border chain.

        Args:
            ch_s (List[Chain]): list of chains
            nodes_s (List[Node]): list of nodes

        Returns:
            None
        """
        from ..analysis.connect_chains import extract_border_chain_from_list

        self.border_chain, self.ch_s_without_border, self.nodes_s_without_border = (
            extract_border_chain_from_list(ch_s, nodes_s)
        )

    def get_iteration_parameters(self, counter: int) -> dict:
        """
        Get the parameters for the next iteration.

        Args:
            counter (int): iteration counter

        Returns:
            Dict: parameters for the next iteration
        """
        iteration_params = {
            "th_radial_tolerance": self.params["th_radial_tolerance"][counter],
            "th_distribution_size": self.params["th_distribution_size"][counter],
            "neighbourhood_size": self.params["neighbourhood_size"][counter],
            "th_regular_derivative": self.params["th_regular_derivative"][counter],
            "derivative_from_center": self.params["derivative_from_center"][counter],
            "l_ch_s": (
                self.ch_s_without_border
                if counter < self.iterations - 1
                else self.ch_s_without_border + [self.border_chain]
            ),
            "l_nodes_s": (
                self.nodes_s_without_border
                if counter < self.iterations - 1
                else self.nodes_s_without_border + self.border_chain.l_nodes
            ),
        }

        if counter == self.iterations - 1:
            self.border_chain.change_id(len(self.ch_s_without_border))

        return iteration_params

    def update_list_for_next_iteration(
        self, ch_c: List[Chain], nodes_c: List[Node]
    ) -> None:
        """
        Update the list of chains and nodes for the next iteration.

        Args:
            ch_c (List[Chain]): list of chains
            nodes_c (List[Node]): list of nodes

        Returns:
            None
        """
        self.ch_s_without_border, self.nodes_s_without_border = ch_c, nodes_c



from typing import List, Optional
import numpy as np
import cv2

from ..geometry.node import Node
from ..geometry.chain import Chain, EndPoints, TypeChains
from ..geometry.geometry_utils import euclidean_distance_between_nodes
from ..visualization.drawing import Color, Drawing
from ..analysis.interpolation_nodes import (
    generate_nodes_list_between_two_radial_distances,
)


class VirtualBandGenerator:
    """
    Class for generating a virtual band between two chains.

    It also incorporates a method to check if a chain is inside the band.
    """

    DOWN = 0
    INSIDE = 1
    UP = 2

    def __init__(
        self,
        l_nodes: List[Node],
        ch_j: Chain,
        ch_k: Chain,
        endpoint: EndPoints,
        ch_i: Optional[Chain] = None,
        band_width: Optional[float] = None,
        debug: bool = False,
        inf_band: Optional[List[Node]] = None,
        sup_band: Optional[List[Node]] = None,
        domain: Optional[List[float]] = None,
    ):
        """
        Initializes an VirtualBandGenerator instance.

        Args:
            l_nodes (List[Node]): List of interpolated nodes between the two chains plus the two endpoints.
            ch_j (Chain): First chain to compute the band.
            ch_k (Chain): Second chain to compute the band.
            endpoint (EndPoints): Endpoint of ch_j.
            ch_i (Optional[Chain]): Support chain of ch_j and ch_k.
            band_width (Optional[float]): Band width as a percentage of the distance between ch_j and support chain (ch_i).
            debug (bool): If True, debugging information will be used.
            inf_band (Optional[List[Node]]): Lower band limit nodes.
            sup_band (Optional[List[Node]]): Upper band limit nodes.
            domain (Optional[List[float]]): Domain of interpolation.
        """
        if band_width is None:
            band_width = 0.05 if ch_i and ch_i.type == TypeChains.center else 0.1

        self.l_nodes = l_nodes
        self.ch_j = ch_j
        self.ch_k = ch_k
        self.endpoint = endpoint
        self.ch_i = ch_i
        params = {
            "y": self.ch_j.center[1],
            "x": self.ch_j.center[0],
            "angle": 0,
            "radial_distance": 0,
            "chain_id": -1,
        }
        self.center = Node(**params)

        ext1 = self.ch_j.extB if endpoint == EndPoints.B else self.ch_j.extA
        ext1_support = (
            self.ch_i.get_node_by_angle(ext1.angle)
            if self.ch_i is not None
            else self.center
        )
        ext2 = self.ch_k.extB if endpoint == EndPoints.A else self.ch_k.extA
        ext2_support = (
            self.ch_i.get_node_by_angle(ext2.angle)
            if self.ch_i is not None
            else self.center
        )
        delta_r1 = euclidean_distance_between_nodes(ext1, ext1_support)
        delta_r2 = euclidean_distance_between_nodes(ext2, ext2_support)
        self.inf_cand = delta_r2 * (1 - band_width)
        self.sup_cand = delta_r2 * (1 + band_width)
        self.inf_orig = delta_r1 * (1 - band_width)
        self.sup_orig = delta_r1 * (1 + band_width)

        if not debug:
            self.generate_band()
        else:
            self.inf_band = inf_band
            self.sup_band = sup_band
            self.interpolation_domain = domain

    def generate_band_limit(self, r2: float, r1: float, total_nodes: int) -> List[Node]:
        """
        Generates the band limit nodes between two radial distances.

        Args:
            r2 (float): Second radial distance.
            r1 (float): First radial distance.
            total_nodes (int): Total number of nodes.

        Returns:
            List[Node]: List of generated nodes for the band limit.
        """
        interpolation_domain = [node.angle for node in self.l_nodes]
        endpoint_cad2 = self.l_nodes[-1]
        support_node2 = (
            self.ch_i.get_node_by_angle(endpoint_cad2.angle)
            if self.ch_i is not None
            else self.center
        )
        sign = (
            -1 if support_node2.radial_distance > endpoint_cad2.radial_distance else +1
        )
        generated_dots = generate_nodes_list_between_two_radial_distances(
            r2,
            r1,
            total_nodes,
            interpolation_domain,
            self.ch_k.center,
            sign,
            self.ch_i,
            self.ch_k,
        )
        self.interpolation_domain = interpolation_domain

        return generated_dots

    def generate_band(self) -> None:
        """
        Generates the lower and upper bands based on the original and candidate radial distances.
        """
        total_nodes = len(self.l_nodes)
        r1 = self.sup_orig
        r2 = self.sup_cand
        self.sup_band = self.generate_band_limit(r2, r1, total_nodes)
        r1 = self.inf_orig
        r2 = self.inf_cand
        self.inf_band = self.generate_band_limit(r2, r1, total_nodes)

    @staticmethod
    def mean_radial_in_node_list(node_list: List[Node]) -> float:
        """
        Computes the mean radial distance in a list of nodes.

        Args:
            node_list (List[Node]): List of nodes.

        Returns:
            float: Mean radial distance.
        """
        return np.mean([node.radial_distance for node in node_list])

    def is_dot_in_band(self, node: Node) -> int:
        """
        Determines the relative position of a node with respect to the band.

        Args:
            node (Node): Node to check.

        Returns:
            int: Relative position (DOWN, INSIDE, or UP).
        """
        inf_mean_radii = self.mean_radial_in_node_list(self.inf_band)
        sup_mean_radii = self.mean_radial_in_node_list(self.sup_band)
        inner_band = self.inf_band if inf_mean_radii < sup_mean_radii else self.sup_band
        outer_band = self.sup_band if inf_mean_radii < sup_mean_radii else self.inf_band
        lowest = [n for n in inner_band if n.angle == node.angle][0]
        highest = [n for n in outer_band if n.angle == node.angle][0]

        if node.radial_distance <= lowest.radial_distance:
            relative_position = VirtualBandGenerator.DOWN
        elif highest.radial_distance >= node.radial_distance >= lowest.radial_distance:
            relative_position = VirtualBandGenerator.INSIDE
        else:
            relative_position = VirtualBandGenerator.UP

        return relative_position

    def is_chain_in_band(self, chain: Chain) -> bool:
        """
        Checks if a chain is inside the band.

        Args:
            chain (Chain): Chain to check.

        Returns:
            bool: True if the chain is inside the band, False otherwise.
        """
        node_chain_in_interval = [
            node for node in chain.l_nodes if node.angle in self.interpolation_domain
        ]
        prev_status = None

        for node in node_chain_in_interval:
            res = self.is_dot_in_band(node)
            if res == VirtualBandGenerator.INSIDE:
                return True
            if prev_status is not None and prev_status != res:
                return True
            prev_status = res

        return False

    def generate_chain_from_node_list(self, l_node: List[Node]) -> Chain:
        """
        Generates a chain from a list of nodes.

        Args:
            l_node (List[Node]): List of nodes.

        Returns:
            Chain: Generated chain.
        """
        chain = Chain(
            chain_id=l_node[0].chain_id,
            center=self.ch_j.center,
            img_height=self.ch_j.img_height,
            img_width=self.ch_j.img_width,
            nr=self.ch_j.nr,
        )
        chain.add_nodes_list(l_node)

        return chain

    def draw_band(self, img: np.ndarray, overlapping_chain: List[Chain]) -> np.ndarray:
        """
        Draws the band and the overlapping chains on an image.

        Args:
            img (np.ndarray): Image to draw on.
            overlapping_chain (List[Chain]): List of overlapping chains.

        Returns:
            np.ndarray: Image with the band and chains drawn.
        """
        img = Drawing.chain(
            self.generate_chain_from_node_list(self.inf_band), img, color=Color.orange
        )
        img = Drawing.chain(
            self.generate_chain_from_node_list(self.sup_band), img, color=Color.maroon
        )
        img = Drawing.chain(self.ch_j, img, color=Color.blue)
        img = Drawing.chain(self.ch_k, img, color=Color.yellow)

        if self.ch_i is not None:
            img = Drawing.chain(self.ch_i, img, color=Color.red)
        for chain in overlapping_chain:
            img = Drawing.chain(chain, img, color=Color.purple)

        return img



from typing import List, Optional

import cv2
import numpy as np

from ..visualization.color import Color
from ..geometry.curve import Curve
from ..config import config

DELIMITE_CURVE_ROW = np.array([-1, -1])


def normalized_row_matrix(matrix: np.ndarray) -> np.ndarray:
    """
    Normalizes each row of a matrix.

    Args:
        matrix (np.ndarray): Input matrix.

    Returns:
        np.ndarray: Matrix with rows normalized.
    """
    epsilon = 1e-10  # Small value to avoid division by zero
    sqrt = np.sqrt((matrix**2).sum(axis=1)) + epsilon
    normalized_array = matrix / sqrt[:, np.newaxis]
    return normalized_array


def erosion(erosion_size: int, src: np.ndarray) -> np.ndarray:
    """
    Applies erosion to an image.

    Args:
        erosion_size (int): Size of the erosion kernel.
        src (np.ndarray): Source image.

    Returns:
        np.ndarray: Eroded image.
    """
    erosion_shape = cv2.MORPH_CROSS
    element = cv2.getStructuringElement(
        erosion_shape,
        (2 * erosion_size + 1, 2 * erosion_size + 1),
        (erosion_size, erosion_size),
    )
    erosion_dst = cv2.erode(src, element)

    return erosion_dst


def dilatation(dilatation_size: int, src: np.ndarray) -> np.ndarray:
    """
    Applies dilation to an image.

    Args:
        dilatation_size (int): Size of the dilation kernel.
        src (np.ndarray): Source image.

    Returns:
        np.ndarray: Dilated image.
    """
    dilation_shape = cv2.MORPH_RECT
    element = cv2.getStructuringElement(
        dilation_shape,
        (2 * dilatation_size + 1, 2 * dilatation_size + 1),
        (dilatation_size, dilatation_size),
    )
    dilatation_dst = cv2.dilate(src, element)

    return dilatation_dst


def mask_background(img: np.ndarray) -> np.ndarray:
    """
    Creates a mask for the background of the image.

    Args:
        img (np.ndarray): Input image.

    Returns:
        np.ndarray: Mask of the background.
    """
    mask = np.zeros((img.shape[0], img.shape[1]), dtype=np.uint8)
    mask[img[:, :] == Color.gray_white] = Color.gray_white

    return mask


def blur(img: np.ndarray, blur_size: int = 11) -> np.ndarray:
    """
    Applies Gaussian blur to an image.

    Args:
        img (np.ndarray): Input image.
        blur_size (int): Size of the Gaussian kernel.

    Returns:
        np.ndarray: Blurred image.
    """
    return cv2.GaussianBlur(img, (blur_size, blur_size), 0)


def thresholding(mask: np.ndarray, threshold: int = 0) -> np.ndarray:
    """
    Applies thresholding to a mask.

    Args:
        mask (np.ndarray): Input mask.
        threshold (int): Threshold value.

    Returns:
        np.ndarray: Thresholded mask.
    """
    mask = np.where(mask > threshold, Color.gray_white, 0).astype(np.uint8)

    return mask


def padding_mask(mask: np.ndarray, pad: int = 3) -> np.ndarray:
    """
    Pads the mask with a constant value.

    Args:
        mask (np.ndarray): Input mask.
        pad (int): Padding size.

    Returns:
        np.ndarray: Padded mask.
    """
    mask = np.pad(
        mask, ((pad, pad), (pad, pad)), mode="constant", constant_values=255
    ).astype(np.uint8)
    return mask


def find_border_contour(mask: np.ndarray, img: np.ndarray) -> Optional[np.ndarray]:
    """
    Finds the contour of the border of the disk.

    Args:
        mask (np.ndarray): Background mask.
        img (np.ndarray): Disk image.

    Returns:
        Optional[np.ndarray]: Border contour coordinates, or None if not found.
    """
    contours, _ = cv2.findContours(mask, cv2.RETR_TREE, cv2.CHAIN_APPROX_NONE)

    # 1.0 Perimeter of the image. Used to discard contours similar in size to the image perimeter.
    perimeter_image = 2 * img.shape[0] + 2 * img.shape[1]
    # 1.1 Error threshold. Discard contours similar in size to the image perimeter (within 10%).
    error_threshold = 0.1 * perimeter_image

    # 2.0 Approximate area of the wood cross-section disk.
    area_image = img.shape[0] * img.shape[1]
    approximate_disk_area = area_image / 2

    # 3.0 Find the contour closest in area to the approximate disk area.
    max_cont = None
    area_difference_min = np.inf
    for c in contours:
        contour_area = cv2.contourArea(c)
        contour_perimeter = cv2.arcLength(c, True)
        if np.abs(contour_perimeter - perimeter_image) < error_threshold:
            # Contour similar to image perimeter. Discard it.
            continue

        area_difference = np.abs(contour_area - approximate_disk_area)
        if area_difference < area_difference_min:
            max_cont = c.reshape((-1, 2))
            area_difference_min = area_difference

    return max_cont


def contour_to_curve(contour: np.ndarray, name: int) -> Curve:
    """
    Converts a contour to a Curve object.

    Args:
        contour (np.ndarray): Contour points.
        name (int): Identifier for the curve.

    Returns:
        Curve: Curve object created from the contour.
    """
    curve = Curve(contour.tolist(), name)

    return curve


def get_border_curve(img: np.ndarray, l_ch_f: List[Curve]) -> Curve:
    """
    Gets the border curve of the disk image. Implements Algorithm 5 in the supplementary material.

    Args:
        img (np.ndarray): Segmented grayscale image.
        l_ch_f (List[Curve]): List of curves.

    Returns:
        Curve: Border curve object.

    Raises:
        ValueError: If border contour cannot be found.
    """
    mask = mask_background(img)
    mask = blur(mask)
    mask = thresholding(mask)
    mask = padding_mask(mask)
    border_contour = find_border_contour(mask, img)

    if border_contour is None:
        raise ValueError("Border contour could not be found.")

    border_curve = contour_to_curve(border_contour, len(l_ch_f))

    return border_curve


def change_reference_axis(ch_e_matrix: np.ndarray) -> np.ndarray:
    """
    Changes the reference axis of the edge matrix to the center (cx, cy).

    Args:
        ch_e_matrix (np.ndarray): Edge matrix.

    Returns:
        np.array: Transformed edge matrix.
    """
    center = [config.cx, config.cy]
    curve_border_index = np.where(ch_e_matrix == DELIMITE_CURVE_ROW)[0]
    X = ch_e_matrix.copy()

    # Change reference axis
    Xb = np.array([[1, 0], [0, 1]]).dot(X.T) + (
        np.array([-1, -1]) * np.array(center, dtype=float)
    ).reshape((-1, 1))

    # Mask delimiting edge row by -1
    Xb[:, curve_border_index] = -1

    return Xb


def convert_masked_pixels_to_curves(
    X_edges_filtered: np.ndarray,
) -> List[Curve]:
    """
    Converts masked pixels into Curve objects.

    Args:
        X_edges_filtered (np.ndarray): Filtered edge matrix.

    Returns:
        List[Curve]: List of Curve objects.
    """
    curve_border_index = np.unique(np.where(X_edges_filtered == DELIMITE_CURVE_ROW)[0])
    start = -1
    ch_f = []

    for end in curve_border_index:
        if end - start > 2:
            pixel_list = X_edges_filtered[start + 1 : end].tolist()
            curve = Curve(pixel_list, len(ch_f))
            ch_f.append(curve)
        start = end

    return ch_f


def get_gradient_vector_for_each_edge_pixel(
    ch_e: np.ndarray, Gx: np.ndarray, Gy: np.ndarray
) -> np.ndarray:
    """
    Gets the gradient vector for each edge pixel.

    Args:
        ch_e (np.ndarray): Edge matrix.
        Gx (np.ndarray): Gradient image in x direction.
        Gy (np.ndarray): Gradient image in y direction.

    Returns:
        np.ndarray: Gradient vectors.
    """
    G = np.vstack(
        (
            Gx[ch_e[:, 1].astype(int), ch_e[:, 0].astype(int)],
            Gy[ch_e[:, 1].astype(int), ch_e[:, 0].astype(int)],
        )
    ).T

    return G


def compute_angle_between_gradient_and_edges(
    Xb_normed: np.ndarray, gradient_normed: np.ndarray
) -> np.ndarray:
    """
    Computes the angle between the gradient and the edges.

    Args:
        Xb_normed (np.ndarray): Normalized edge vectors.
        gradient_normed (np.ndarray): Normalized gradient vectors.

    Returns:
        np.ndarray: Angles in degrees.
    """
    theta = (
        np.arccos(np.clip((gradient_normed * Xb_normed).sum(axis=1), -1.0, 1.0))
        * 180
        / np.pi
    )

    return theta


def filter_edges_by_threshold(m_ch_e: np.ndarray, theta: np.ndarray) -> np.ndarray:
    """
    Filters edges by a threshold angle.

    Args:
        m_ch_e (np.ndarray): Edge matrix.
        theta (np.ndarray): Angles between gradient and edges.
        alpha (float): Threshold angle.

    Returns:
        np.ndarray: Filtered edge matrix.
    """
    X_edges_filtered = m_ch_e.copy()
    X_edges_filtered[theta >= config.alpha] = -1

    return X_edges_filtered


def filter_edges(
    devernay_edges: np.ndarray,
    gradient_x_img: np.ndarray,
    gradient_y_img: np.ndarray,
    img_pre: np.ndarray,
) -> List[Curve]:
    """
    Filters edges to keep only the ones forming rings (early wood edges).

    Edge detector finds three types of edges: early wood transitions, latewood transitions, and radial edges produced by
    cracks and fungi. Only early wood edges form the rings. To filter out the others, the collinearity with the ray direction
    is computed and filtered depending on threshold (alpha). Implements Algorithm 4 in the supplementary material.

    Args:
        devernay_edges (np.ndarray): Devernay curves in matrix format.
        gradient_x_img (np.ndarray): Gradient image in x direction.
        gradient_y_img (np.ndarray): Gradient image in y direction.
        img_pre (np.ndarray): Input image.

    Returns:
        List[Curve]: Filtered Devernay curves.
    """
    # Change reference axis
    Xb = change_reference_axis(devernay_edges)

    # Get normalized gradient at each edge
    G = get_gradient_vector_for_each_edge_pixel(
        devernay_edges, gradient_x_img, gradient_y_img
    )

    # Normalize gradient and rays
    Xb_normalized = normalized_row_matrix(Xb.T)
    G_normalized = normalized_row_matrix(G)

    # Compute angle between gradient and edges
    theta = compute_angle_between_gradient_and_edges(Xb_normalized, G_normalized)

    # Filter pixels by threshold
    X_edges_filtered = filter_edges_by_threshold(devernay_edges, theta)

    # Convert masked pixels to Curve objects
    l_ch_f = convert_masked_pixels_to_curves(X_edges_filtered)

    # Add border curve
    border_curve = get_border_curve(img_pre, l_ch_f)

    # Append border curve to the list
    l_ch_f.append(border_curve)

    return l_ch_f



from typing import Any, List, Tuple

import cv2
import numpy as np
from shapely.geometry import Point

from ..geometry.chain import (
    Chain,
    Node,
    TypeChains,
)
from ..visualization.drawing import Drawing
from ..geometry.ray import Ray
from ..geometry.curve import Curve
from ..geometry.geometry_utils import euclidean_distance, get_node_from_list_by_angle
from ..config import config


def build_rays(m: int, n: int, center: Tuple[float, float]) -> List[Ray]:
    """
    Builds a list of rays emanating from the center at evenly spaced angles.

    Args:
        m (int): Height of the image.
        n (int): Width of the image.
        center (Tuple[float, float]): Center coordinates (y, x).

    Returns:
        List[Ray]: List of Ray objects.
    """
    angles_range = np.arange(0, 360, 360 / config.nr)
    radii_list = [Ray(direction, center, m, n) for direction in angles_range]
    return radii_list


def get_coordinates_from_intersection(inter: Any) -> Tuple[float, float]:
    """
    Formats the coordinates from a Shapely intersection object.

    Args:
        inter (Any): Intersection object.

    Returns:
        Tuple[float, float]: Coordinates (y, x) of the intersection point.
    """
    if "MULTI" in inter.wkt:
        inter = inter.geoms[0]

    if isinstance(inter, Point):
        y, x = inter.xy
        y, x = y[0], x[0]
    elif "LINESTRING" in inter.wkt:
        y, x = inter.xy
        y, x = y[1], x[1]
    elif "STRING" in inter.wkt:
        y, x = inter.coords.xy
        y, x = y[0], x[0]
    else:
        raise ValueError("Unsupported intersection type")

    return y, x


def compute_intersection(
    l_rays: List[Ray], curve: Curve, chain_id: int, center: Tuple[float, float]
) -> List[Node]:
    """
    Computes the intersection between rays and a Devernay curve.

    Args:
        l_rays (List[Ray]): List of rays.
        curve (Curve): Devernay curve with a 'geometry' attribute.
        chain_id (int): Chain identifier.
        center (Tuple[float, float]): Center coordinates (y, x).

    Returns:
        List[Node]: List of nodes resulting from the intersections.
    """
    l_curve_nodes = []
    for radii in l_rays:
        inter = radii.geometry.intersection(curve.geometry)
        if not inter.is_empty:
            try:
                y, x = get_coordinates_from_intersection(inter)
            except NotImplementedError:
                continue
            i, j = np.array(y), np.array(x)
            params = {
                "y": i,
                "x": j,
                "angle": int(radii.direction),
                "radial_distance": euclidean_distance([i, j], center),
                "chain_id": chain_id,
            }

            dot = Node(**params)
            if (
                dot not in l_curve_nodes
                and get_node_from_list_by_angle(l_curve_nodes, radii.direction) is None
            ):
                l_curve_nodes.append(dot)

    return l_curve_nodes


def intersections_between_rays_and_devernay_curves(
    l_rays: List[Ray],
    l_curves: List[Curve],
    center: Tuple[float, float],
    height: int,
    width: int,
) -> Tuple[List[Node], List[Chain]]:
    """
    Computes chains by sampling Devernay curves. Sampling is performed by finding the intersection
    between rays and Devernay curves.

    Args:
        center (Tuple[float, float]): Center coordinates (y, x).
        l_rays (List[Ray]): List of rays.
        l_curves (List[Curve]): List of Devernay curves.
        height (int): Height of the image.
        width (int): Width of the image.

    Returns:
        Tuple[List[Node], List[Chain]]: List of nodes and list of chains.
    """
    l_chain, l_nodes = [], []

    for idx, curve in enumerate(l_curves):
        chain_id = len(l_chain)
        l_curve_nodes = compute_intersection(l_rays, curve, chain_id, center)

        if len(l_curve_nodes) < config.min_chain_length:
            continue

        l_nodes += l_curve_nodes
        chain = Chain(
            chain_id, config.nr, center=center, img_height=height, img_width=width
        )
        chain.add_nodes_list(l_curve_nodes)
        l_chain.append(chain)

    # Devernay border curve is the last element of the list l_curves.
    if l_chain:
        l_chain[-1].type = TypeChains.border

    return l_nodes, l_chain


def generate_virtual_center_chain(
    chains_list: List[Chain],
    nodes_list: List[Node],
    height: int,
    width: int,
) -> None:
    """
    Generates a virtual center chain. This chain is used to connect the other chains.

    Args:
        chains_list (List[Chain]): List of chains.
        nodes_list (List[Node]): List of nodes.
        height (int): Height of the image.
        width (int): Width of the image.

    Returns:
        None
    """
    chain_id = len(chains_list)
    center_list = [
        Node(
            x=config.cx,
            y=config.cy,
            angle=angle,
            radial_distance=0,
            chain_id=chain_id,
        )
        for angle in np.arange(0, 360, 360 / config.nr)
    ]
    nodes_list += center_list

    chain = Chain(
        chain_id,
        config.nr,
        center=chains_list[0].center if chains_list else (config.cy, config.cx),
        img_height=height,
        img_width=width,
        type=TypeChains.center,
    )
    chain.add_nodes_list(center_list)

    chains_list.append(chain)

    # Set border chain as the last element of the list
    if len(chains_list) >= 2:
        chains_list[-2].change_id(len(chains_list) - 1)


def draw_ray_curve_and_intersections(
    dots_lists: List[Node],
    rays_list: List[Ray],
    curves_list: List[Curve],
    img_draw: np.ndarray,
    filename: str,
) -> None:
    """
    Draws rays, curves, and intersections on an image and saves it.

    Args:
        dots_lists (List[Node]): List of nodes representing intersections.
        rays_list (List[Ray]): List of rays.
        curves_list (List[Curve]): List of curves.
        img_draw (np.ndarray): Image array to draw on.
        filename (str): Filename to save the image.

    Returns:
        None
    """
    for ray in rays_list:
        img_draw = Drawing.radii(ray, img_draw)

    for curve in curves_list:
        img_draw = Drawing.curve(curve, img_draw)

    for dot in dots_lists:
        img_draw = Drawing.intersection(dot, img_draw)

    cv2.imwrite(filename, img_draw)


def sampling_edges(
    l_ch_f: List[Curve],
    img_pre: np.ndarray,
) -> Tuple[List[Chain], List[Node]]:
    """
    Samples Devernay curves using ray directions. Implements Algorithm 6 in the supplementary material.

    Args:
        l_ch_f (List[Curve]): Devernay curves.
        img_pre (np.ndarray): Input image.

    Returns:
        Tuple[List[Chain], List[Node]]:
            - l_ch_s: Sampled edges curves (list of Chain objects).
            - l_nodes_s: List of nodes.
    """
    height, width = img_pre.shape[:2]
    center = (config.cy, config.cx)

    l_rays = build_rays(height, width, center)
    l_nodes_s, l_ch_s = intersections_between_rays_and_devernay_curves(
        l_rays, l_ch_f, center, height, width
    )
    generate_virtual_center_chain(l_ch_s, l_nodes_s, height, width)

    # Debug purposes, not illustrated in the paper
    if config.debug:
        img_draw = np.zeros((img_pre.shape[0], img_pre.shape[1], 3), dtype=np.uint8)
        draw_ray_curve_and_intersections(
            l_nodes_s,
            l_rays,
            l_ch_f,
            img_draw,
            f"{config.output_dir}/dots_curve_and_rays.png",
        )

    return l_ch_s, l_nodes_s



import numpy as np
from pathlib import Path
from typing import List, Optional, Tuple

from ..geometry.geometry_utils import (
    copy_chain,
    euclidean_distance_between_nodes,
    get_chain_from_list_by_id,
    get_chains_within_angle,
)
from ..geometry.chain import Chain, EndPoints
from ..geometry.node import Node
from ..analysis.interpolation_nodes import domain_interpolation
from ..analysis.chain_analysis_tools import (
    exist_chain_overlapping,
)
from ..config import config


class ChainSystemManager:
    def __init__(
        self,
        l_ch: List[Chain],
        l_nodes: List[Node],
        M: np.ndarray,
        nr: int = 360,
        th_radial_tolerance: float = 0.1,
        th_distribution_size: int = 2,
        th_regular_derivative: float = 1.5,
        neighbourhood_size: int = 45,
        derivative_from_center: bool = False,
        debug: bool = False,
        counter: int = 0,
        save: Optional[str] = None,
        img: Optional[np.ndarray] = None,
    ) -> None:
        """
        Initialize the ChainSystemManager with the given parameters.

        Args:
            l_ch (List[Chain]): List of chains.
            l_nodes (List[Node]): List of nodes.
            M (np.ndarray): Matrix representing the system.
            cy (int): Y-coordinate of the center.
            cx (int): X-coordinate of the center.
            nr (int, optional): Number of radial divisions. Default is 360.
            th_radial_tolerance (float, optional): Radial tolerance threshold. Default is 0.1.
            th_distribution_size (int, optional): Distribution size threshold. Default is 2.
            th_regular_derivative (float, optional): Regular derivative threshold. Default is 1.5.
            neighbourhood_size (int, optional): Size of the neighbourhood. Default is 45.
            derivative_from_center (bool, optional): Whether to calculate derivative from center. Default is False.
            debug (bool, optional): Enable debugging. Default is False.
            counter (int, optional): Counter value. Default is 0.
            save (Optional[str], optional): Path to save the results. Default is None.
            img (Optional[np.ndarray], optional): Image array. Default is None.

        Returns:
            None
        """
        # initialization
        self.l_nodes_s = l_nodes
        self.l_ch_s = l_ch
        self.__sort_chain_list_and_update_relative_position()

        # system parameters
        self.nr = nr
        self.derivative_from_center = derivative_from_center
        self.th_distribution_size = th_distribution_size
        self.debug = debug
        self.neighbourhood_size = neighbourhood_size
        self.M = M
        self.center = [config.cy, config.cx]
        self.img = img
        self.height = img.shape[0]
        self.width = img.shape[1]
        self.next_chain_index = 0
        self.iterations_since_last_change = 0
        self.th_radial_tolerance = th_radial_tolerance
        self.label = "system_status"
        self.counter = counter
        self.th_regular_derivative = th_regular_derivative
        self.path = save

        if self.path is not None and self.debug:
            Path(self.path).mkdir(exist_ok=True)

    def get_common_chain_to_both_borders(self, chain: Chain) -> Optional[Chain]:
        """
        Args:
            chain (Chain): The chain to check for common chains to both borders.

        Returns:
            Optional[Chain]: The common chain to both borders if found, otherwise None.
        """
        chain_angle_domain = chain.get_dot_angle_values()
        angles_where_there_is_no_nodes = [
            angle
            for angle in np.arange(0, 360, 360 / self.nr)
            if angle not in chain_angle_domain
        ]
        angles_where_there_is_no_nodes += [chain.extA.angle, chain.extB.angle]
        chains_where_there_is_no_nodes = []
        for ch_i in self.l_ch_s:
            ch_i_angles = ch_i.get_dot_angle_values()
            if np.intersect1d(ch_i_angles, angles_where_there_is_no_nodes).shape[
                0
            ] == len(angles_where_there_is_no_nodes):
                chains_where_there_is_no_nodes.append(ch_i)

        if len(chains_where_there_is_no_nodes) == 0:
            return None

        nodes_in_ray_a = [
            cad.get_node_by_angle(chain.extA.angle)
            for cad in chains_where_there_is_no_nodes
        ]
        nodes_in_ray_a.sort(
            key=lambda x: euclidean_distance_between_nodes(x, chain.extA)
        )
        id_closest = nodes_in_ray_a[0].chain_id

        return get_chain_from_list_by_id(self.l_ch_s, id_closest)

    def compute_all_elements_needed_to_check_if_exist_chain_overlapping(
        self, chain: Chain
    ) -> Tuple[Optional[Chain], List[Node], EndPoints]:
        """
        Args:
            chain (Chain): The chain to check for overlapping.

        Returns:
            Tuple[Optional[Chain], List[Node], EndPoints]: The common chain, interpolated nodes plus endpoints, and endpoint type.
        """
        ch_i = self.get_common_chain_to_both_borders(chain)
        ch_j_endpoint_node = chain.extB
        ch_k_endpoint_node = chain.extA
        ch_j_endpoint_type = EndPoints.B

        interpolated_nodes = []  # domain interpolation function output
        chain_copy = copy_chain(chain)
        domain_interpolation(
            ch_i,
            ch_j_endpoint_node,
            ch_k_endpoint_node,
            ch_j_endpoint_type,
            chain_copy,
            interpolated_nodes,
        )
        interpolated_nodes_plus_endpoints = (
            [ch_j_endpoint_node] + interpolated_nodes + [ch_k_endpoint_node]
        )

        return ch_i, interpolated_nodes_plus_endpoints, ch_j_endpoint_type

    def fill_chain_if_there_is_no_overlapping(self, chain: Chain) -> None:
        """
        Algorithm 10 in the supplementary material. Complete chain if there is no overlapping.

        Args:
            chain (Chain): Chain to be completed if conditions are met.

        Returns:
            None
        """
        # Check if chain is too small or is full. If it is, return.
        threshold = 0.9
        if chain.size >= chain.nr or chain.size < threshold * chain.nr:
            return

        ch_i, l_nodes, endpoint_type = (
            self.compute_all_elements_needed_to_check_if_exist_chain_overlapping(chain)
        )

        # Algorithm 13 in the paper. Check if there is an overlapping chain.
        exist_chain = exist_chain_overlapping(
            self.l_ch_s, l_nodes, chain, chain, endpoint_type, ch_i
        )

        if exist_chain:
            return

        self.add_nodes_list_to_system(chain, l_nodes)

        return

    def continue_in_loop(self) -> bool:
        """
        Args:
            None

        Returns:
            bool: Whether the algorithm should continue in the loop.
        """
        # If iteration_sin_last_chain is equal to the number of chains, the algorithm stops. This means that
        # all chains have been iterated at least one time and no chains have been connected or interpolated.
        return self.iterations_since_last_change < len(self.l_ch_s)

    def get_next_chain(self) -> Chain:
        """
        Algorithm 9 in the supplementary material. Get next chain to be processed.

        Args:
            None

        Returns:
            Chain: Next supported chain.
        """
        ch_i = self.l_ch_s[self.next_chain_index]

        self.size_l_chain_init = len(self.l_ch_s)

        #  Algorithm 12 in the paper
        self.fill_chain_if_there_is_no_overlapping(ch_i)

        return ch_i

    def is_new_dot_valid(self, new_dot: Node) -> bool:
        """
        Args:
            new_dot (Node): The new dot to validate.

        Returns:
            bool: Whether the new dot is valid.
        """
        if new_dot in self.l_nodes_s:
            return False
        if (
            new_dot.x >= self.height
            or new_dot.y >= self.width
            or new_dot.x < 0
            or new_dot.y < 0
        ):
            return False

        return True

    def update_chain_neighbourhood(
        self, l_chains_to_update_neighborhood: List[Chain]
    ) -> None:
        """
        Args:
            l_chains_to_update_neighborhood (List[Chain]): List of chains to update the neighborhood.

        Returns:
            None
        """
        from ..analysis.connect_chains import get_inward_and_outward_visible_chains

        dummy_chain = None
        for chain_p in l_chains_to_update_neighborhood:
            border = EndPoints.A
            inward_chain, outward_chain, dot_border = (
                get_inward_and_outward_visible_chains(self.l_ch_s, chain_p, border)
            )

            chain_p.A_outward = (
                outward_chain if outward_chain is not None else dummy_chain
            )
            chain_p.A_inward = inward_chain if inward_chain is not None else dummy_chain
            border = EndPoints.B
            inward_chain, outward_chain, dot_border = (
                get_inward_and_outward_visible_chains(self.l_ch_s, chain_p, border)
            )
            chain_p.B_outward = (
                outward_chain if outward_chain is not None else dummy_chain
            )
            chain_p.B_inward = inward_chain if inward_chain is not None else dummy_chain

        return

    def add_nodes_list_to_system(self, chain: Chain, l_nodes: List[Node]) -> None:
        """
        Args:
            chain (Chain): The chain to which nodes will be added.
            l_nodes (List[Node]): List of nodes to be added to the system.

        Returns:
            None
        """
        processed_node_list = []
        for new_dot in l_nodes:
            if chain.id != new_dot.chain_id:
                raise

            processed_node_list.append(new_dot)
            if new_dot in self.l_nodes_s:
                continue

            self.l_nodes_s.append(new_dot)
            # 1.0 Update ch_i list intersection
            chain_id_intersecting, chains_over_radial_direction = (
                self._chains_id_over_radial_direction(new_dot.angle)
            )
            self.M[chain.id, chain_id_intersecting] = 1
            self.M[chain_id_intersecting, chain.id] = 1

            # 2.0 Update boundary chains above and below.
            dots_over_direction = [
                dot
                for chain in chains_over_radial_direction
                for dot in chain.l_nodes
                if dot.angle == new_dot.angle
            ]
            dots_over_direction.append(new_dot)
            dots_over_direction.sort(key=lambda x: x.radial_distance)
            idx_new_dot = dots_over_direction.index(new_dot)

            up_dot = (
                dots_over_direction[idx_new_dot + 1]
                if idx_new_dot < len(dots_over_direction) - 1
                else None
            )
            if up_dot is not None:
                up_chain = get_chain_from_list_by_id(
                    chain_list=chains_over_radial_direction, chain_id=up_dot.chain_id
                )
                if up_dot == up_chain.extA:
                    up_chain.A_inward = chain
                elif up_dot == up_chain.extB:
                    up_chain.B_inward = chain

            down_dot = dots_over_direction[idx_new_dot - 1] if idx_new_dot > 0 else None
            if down_dot is not None:
                down_chain = get_chain_from_list_by_id(
                    chain_list=chains_over_radial_direction, chain_id=down_dot.chain_id
                )
                if down_dot == down_chain.extA:
                    down_chain.A_outward = chain
                elif down_dot == down_chain.extB:
                    down_chain.B_outward = chain

        change_border = chain.add_nodes_list(processed_node_list)
        self.update_chain_neighbourhood([chain])

    @staticmethod
    def get_next_chain_index_in_list(
        chains_list: List[Chain], support_chain: Chain
    ) -> int:
        """
        Args:
            chains_list (List[Chain]): List of chains.
            support_chain (Chain): The support chain to find the next index for.

        Returns:
            int: The index of the next chain in the list.
        """
        return (chains_list.index(support_chain) + 1) % len(chains_list)

    def update_system_status(
        self, ch_i: Chain, l_s_outward: List[Chain], l_s_inward: List[Chain]
    ) -> int:
        """
        Algorithm 8 in the supplementary material. Update the system state after the iteration.

        Args:
            ch_i (Chain): Support chain in current iteration.
            l_s_outward (List[Chain]): Outward list of chains.
            l_s_inward (List[Chain]): Inward list of chains.

        Returns:
            int: Index of the next support chain to be processed.
        """
        # self.l_ch_s is an attribute of the self object.
        if self._system_status_change():
            self.l_ch_s.sort(key=lambda x: x.size, reverse=True)
            # Variable used to check if system status has changed in current iteration. If it has (variable is set to 0),
            # the algorithm continues one more iteration. If it has not (variable is increased by 1). Variable is used as
            # exit condition in the while loop (Algorithm 2, line 3, continue_in_loop method).
            self.iterations_since_last_change = 0

            l_current_iteration = [ch_i] + l_s_outward + l_s_inward

            l_current_iteration.sort(key=lambda x: x.size, reverse=True)

            longest_chain = l_current_iteration[0]

            if longest_chain.id == ch_i.id:
                next_chain_index = self.get_next_chain_index_in_list(self.l_ch_s, ch_i)
            else:
                next_chain_index = self.l_ch_s.index(longest_chain)

        else:
            next_chain_index = self.get_next_chain_index_in_list(self.l_ch_s, ch_i)
            # System status has not changed. Increase variable by 1.
            self.iterations_since_last_change += 1

        self.next_chain_index = next_chain_index

        return 0

    def _chains_id_over_radial_direction(
        self, angle: float
    ) -> Tuple[List[int], List[Chain]]:
        """
        Args:
            angle (float): The angle to find chains over radial direction.

        Returns:
            Tuple[List[int], List[Chain]]: List of chain IDs and chains in radial direction.
        """
        chains_in_radial_direction = get_chains_within_angle(angle, self.l_ch_s)
        chains_id_over_radial_direction = [cad.id for cad in chains_in_radial_direction]

        return chains_id_over_radial_direction, chains_in_radial_direction

    def __sort_chain_list_and_update_relative_position(self) -> None:
        """
        Args:
            None

        Returns:
            None
        """
        self.l_ch_s = sorted(self.l_ch_s, key=lambda x: x.size, reverse=True)
        self.update_chain_neighbourhood(self.l_ch_s)

    def _system_status_change(self) -> bool:
        """
        Args:
            None

        Returns:
            bool: Whether the system status has changed.
        """
        self.chain_size_at_the_end_of_iteration = len(self.l_ch_s)

        return self.size_l_chain_init > self.chain_size_at_the_end_of_iteration



from enum import Enum
from typing import List, Optional, Tuple, Any

import numpy as np

from ..geometry.node import Node


class TypeChains(Enum):
    center = 0
    normal = 1
    border = 2


class ClockDirection(Enum):
    clockwise = 0
    anti_clockwise = 1


class EndPoints(Enum):
    A = 0
    B = 1


class ChainLocation(Enum):
    inwards = 0
    outwards = 1


class Chain:
    def __init__(
        self,
        chain_id: int,
        nr: int,
        center: Tuple[float, float],
        img_height: int,
        img_width: int,
        type: TypeChains = TypeChains.normal,
        A_outward: Optional[Node] = None,
        A_inward: Optional[Node] = None,
        B_outward: Optional[Node] = None,
        B_inward: Optional[Node] = None,
    ) -> None:
        """
        Represents a chain of nodes.

        Args:
            chain_id (int): Identifier for the chain.
            nr (int): Number of rays.
            center (Tuple[float, float]): Center coordinates (x, y).
            img_height (int): Image height.
            img_width (int): Image width.
            type (TypeChains): Type of the chain.
            A_outward (Optional[Node]): Outward node at endpoint A.
            A_inward (Optional[Node]): Inward node at endpoint A.
            B_outward (Optional[Node]): Outward node at endpoint B.
            B_inward (Optional[Node]): Inward node at endpoint B.

        Returns:
            None
        """
        self.l_nodes: List[Node] = []
        self.id = chain_id
        self.size = 0
        self.nr = nr
        self.extA: Optional[Node] = None
        self.extB: Optional[Node] = None
        self.type = type
        self.A_outward = A_outward
        self.A_inward = A_inward
        self.B_outward = B_outward
        self.B_inward = B_inward

        # Less important attributes
        self.center = center  # center of the disk (pith pixel location)
        self.img_height = img_height  # image height
        self.img_width = img_width  # image width
        self.label_id = chain_id  # debug purpose

    def __eq__(self, other: Any) -> bool:
        """
        Compares two chains for equality.

        Args:
            other (Any): Object to compare to.

        Returns:
            bool: True if the chains are equal, False otherwise.
        """
        return (
            isinstance(other, Chain) and self.id == other.id and self.size == other.size
        )

    def is_closed(self, threshold: float = 0.95) -> bool:
        """
        Checks if the chain is closed based on the threshold.

        Args:
            threshold (float): Threshold ratio for being considered closed.

        Returns:
            bool: True if the chain is closed, False otherwise.
        """
        return len(self.l_nodes) >= threshold * self.nr

    def sort_dots(
        self, direction: ClockDirection = ClockDirection.clockwise
    ) -> List[Node]:
        """
        Returns the nodes sorted in the specified direction.

        Args:
            direction (ClockDirection): Direction to sort the nodes.

        Returns:
            List[Node]: Sorted list of nodes.
        """
        return (
            self.clockwise_sorted_dots
            if direction == ClockDirection.clockwise
            else self.clockwise_sorted_dots[::-1]
        )

    def _sort_dots(
        self, direction: ClockDirection = ClockDirection.clockwise
    ) -> List[Node]:
        """
        Sorts the nodes in the chain in the specified direction.

        Args:
            direction (ClockDirection): Direction to sort the nodes.

        Returns:
            List[Node]: Sorted list of nodes.
        """
        clock_wise_sorted_dots = []
        step = 360 / self.nr
        angle_k = (
            self.extB.angle
            if direction == ClockDirection.clockwise
            else self.extA.angle
        )
        while len(clock_wise_sorted_dots) < self.size:
            dot = self.get_node_by_angle(angle_k)
            assert dot is not None
            assert dot.chain_id == self.id
            clock_wise_sorted_dots.append(dot)
            angle_k = (
                (angle_k - step) % 360
                if direction == ClockDirection.clockwise
                else (angle_k + step) % 360
            )
        return clock_wise_sorted_dots

    def __repr__(self) -> str:
        """
        Returns a string representation of the chain.

        Args:
            None

        Returns:
            str: String representation of the chain.
        """
        return f"(id_l:{self.label_id},id:{self.id}, size {self.size}"

    def __find_endpoints(self) -> bool:
        """
        Finds the endpoints of the chain.

        Args:
            None

        Returns:
            bool: True if endpoints changed, False otherwise.
        """
        diff = np.zeros(self.size)
        extA_init = self.extA
        extB_init = self.extB
        self.l_nodes.sort(key=lambda x: x.angle, reverse=False)
        diff[0] = (self.l_nodes[0].angle + 360 - self.l_nodes[-1].angle) % 360

        for i in range(1, self.size):
            diff[i] = self.l_nodes[i].angle - self.l_nodes[i - 1].angle

        border1 = diff.argmax()
        if border1 == 0:
            border2 = diff.shape[0] - 1
        else:
            border2 = border1 - 1

        self.extAind = border1
        self.extBind = border2

        change_border = (
            True
            if (extA_init is None or extB_init is None)
            or (
                extA_init != self.l_nodes[border1] or extB_init != self.l_nodes[border2]
            )
            else False
        )
        self.extA = self.l_nodes[border1]
        self.extB = self.l_nodes[border2]

        return change_border

    def add_nodes_list(self, l_nodes: List[Node]) -> bool:
        """
        Adds a list of nodes to the chain.

        Args:
            l_nodes (List[Node]): List of nodes to add.

        Returns:
            bool: True if endpoints changed after adding nodes, False otherwise.
        """
        self.l_nodes += l_nodes
        change_border = self.update()
        return change_border

    def update(self) -> bool:
        """
        Updates the chain after adding nodes.

        Args:
            None

        Returns:
            bool: True if endpoints changed, False otherwise.
        """
        self.size = len(self.l_nodes)
        if self.size > 1:
            change_border = self.__find_endpoints()
            self.clockwise_sorted_dots = self._sort_dots()
        else:
            raise Exception("Chain must have at least two nodes.")

        return change_border

    def get_nodes_coordinates(self) -> Tuple[np.ndarray, np.ndarray]:
        """
        Gets the coordinates of the nodes in the chain.

        Args:
            None

        Returns:
            Tuple[np.ndarray, np.ndarray]: Arrays of x and y coordinates.
        """
        x = [dot.x for dot in self.l_nodes]
        y = [dot.y for dot in self.l_nodes]
        x_rot = np.roll(x, -self.extAind)
        y_rot = np.roll(y, -self.extAind)

        return np.array(x_rot), np.array(y_rot)

    def get_dot_angle_values(self) -> List[float]:
        """
        Gets the angle values of the nodes in the chain.

        Args:
            None

        Returns:
            List[float]: List of angle values.
        """
        return [dot.angle for dot in self.l_nodes]

    def get_node_by_angle(self, angle: float) -> Optional[Node]:
        """
        Retrieves a node by its angle.

        Args:
            angle (float): Angle to search for.

        Returns:
            Optional[Node]: The node with the specified angle, or None if not found.
        """
        from ..geometry.geometry_utils import get_node_from_list_by_angle

        return get_node_from_list_by_angle(self.l_nodes, angle)

    def change_id(self, index: int) -> int:
        """
        Changes the ID of the chain and its nodes.

        Args:
            index (int): New ID for the chain.

        Returns:
            int: 0 upon completion.
        """
        for dot in self.l_nodes:
            dot.chain_id = index
        self.id = index
        return 0

    def to_array(
        self,
    ) -> Tuple[np.ndarray, np.ndarray, np.ndarray]:
        """
        Returns nodes coordinates in a numpy array, along with endpoint coordinates.

        Args:
            None

        Returns:
            Tuple[np.ndarray, np.ndarray, np.ndarray]:
                nodes (np.ndarray): Coordinates of nodes.
                c1a (np.array): Coordinates of extA.
                c1b (np.array): Coordinates of extB.
        """
        x1, y1 = self.get_nodes_coordinates()
        nodes = np.vstack((x1, y1)).T
        c1a = np.array([self.extA.x, self.extA.y], dtype=float)
        c1b = np.array([self.extB.x, self.extB.y], dtype=float)

        return nodes.astype(float), c1a, c1b



import logging
import time
from typing import Tuple, List
import numpy as np


from .utils.file_utils import load_image, write_json
from .analyzer import tree_ring_detection
from .config import config
from .utils.results_handler import save_results
from .geometry.curve import Curve
from .geometry.chain import Chain

logger = logging.getLogger(__name__)


def run() -> Tuple[
    np.ndarray,
    np.ndarray,
    np.ndarray,
    List[Curve],
    List[Chain],
    List[Chain],
    List[Chain],
]:
    """
    Main function to run tree ring detection.

    Args:
        None

    Returns:
        Tuple containing:
            - img_in (np.ndarray): Original input image.
            - img_pre (np.ndarray): Preprocessed image.
            - devernay_curves (np.ndarray): Devernay curves in matrix format.
            - devernay_curves_f (List[Curve]): Filtered Devernay curves.
            - devernay_curves_s (List[Chain]): Sampled Devernay curves as Chain objects.
            - devernay_curves_c (List[Chain]): Chain lists after connect stage.
            - devernay_curves_p (List[Chain]): Chain lists after postprocessing stage.
    """
    # Set up logging based on debug setting
    logging.basicConfig(
        level=logging.DEBUG if config.debug else logging.INFO,
        format="%(asctime)s - %(name)s - %(levelname)s - %(message)s",
    )

    try:
        config.log_all_configs()

        logger.info(f"Loading input image: {config.input_image}")

        if config.input_image is None:
            raise ValueError("Input image path cannot be None")

        img_in = load_image(config.input_image)

        logger.info("Running tree ring detection...")

        start_time = time.time()

        (
            img_in,
            img_pre,
            devernay_edges,
            devernay_curves_f,
            devernay_curves_s,
            devernay_curves_c,
            devernay_curves_p,
        ) = tree_ring_detection(img_in)

        exec_time = time.time() - start_time

        if config.save_results:
            logger.info("Saving results...")
            save_results(
                img_in,
                img_pre,
                devernay_edges,
                devernay_curves_f,
                devernay_curves_s,
                devernay_curves_c,
                devernay_curves_p,
            )

            config_path = config.output_dir / "config.json"
            write_json(config.to_dict(), config_path)
            logger.info(f"Saved configuration to {config_path}")

        logger.info(f"Processing completed in {exec_time:.2f} seconds")
        return (
            img_in,
            img_pre,
            devernay_edges,
            devernay_curves_f,
            devernay_curves_s,
            devernay_curves_c,
            devernay_curves_p,
        )

    except Exception as e:
        logger.error(f"Error during processing: {str(e)}", exc_info=True)
        return (
            np.array([]),
            np.array([]),
            np.array([]),
            [],
            [],
            [],
            [],
        )



from ..geometry.chain import Chain


class Set:
    def __init__(self, angular_distance: float, cad: Chain) -> None:
        """
        Set class constructor

        Args:
            angular_distance (float): Angular distance between two points
            cad (Chain): Chain object

        Returns:
            None
        """
        self.distance = angular_distance
        self.cad = cad



import numpy as np
import cv2
from typing import List, Tuple, Dict, Any
from .geometry.curve import Curve
from .geometry.chain import Chain
from .processing.preprocessing import preprocessing
from .detection.canny_devernay_edge_detector import canny_deverney_edge_detector
from .detection.filter_edges import filter_edges
from .processing.sampling import sampling_edges
from .analysis.connect_chains import connect_chains
from .processing.postprocessing import postprocessing
from .config import config


def tree_ring_detection(img_in: cv2.typing.MatLike) -> Tuple[
    np.ndarray,
    np.ndarray,
    np.ndarray,
    List[Curve],
    List[Chain],
    List[Chain],
    List[Chain],
]:
    """
    Delineate tree rings over pine cross-section images. Implements Algorithm 1 from the paper.

    Args:
        img_in (cv2.typing.MatLike): Segmented input image. Background must be white (255,255,255).

    Returns:
        Tuple containing:
            - img_in (np.ndarray): Original input image.
            - img_pre (np.ndarray): Preprocessed image.
            - devernay_curves (np.ndarray): Devernay curves in matrix format.
            - devernay_curves_f (List[Curve]): Filtered Devernay curves.
            - devernay_curves_s (List[Chain]): Sampled Devernay curves as Chain objects.
            - devernay_curves_c (List[Chain]): Chain lists after connect stage.
            - devernay_curves_p (List[Chain]): Chain lists after postprocessing stage.
    """
    img_pre = preprocessing(img_in)

    devernay_result_c = canny_deverney_edge_detector(
        img_pre, sigma=config.sigma, low=config.th_low, high=config.th_high
    )

    devernay_edges, gradient_x_img, gradient_y_img = devernay_result_c

    devernay_curves_f = filter_edges(
        devernay_edges, gradient_x_img, gradient_y_img, img_pre
    )
    devernay_curves_s, l_nodes_s = sampling_edges(devernay_curves_f, img_pre)
    devernay_curves_c, l_nodes_c = connect_chains(devernay_curves_s, img_pre)
    devernay_curves_p = postprocessing(devernay_curves_c, l_nodes_c, img_pre)

    return (
        img_in,
        img_pre,
        devernay_edges,
        devernay_curves_f,
        devernay_curves_s,
        devernay_curves_c,
        devernay_curves_p,
    )



import numpy as np
from typing import List, Tuple, Optional

from ..geometry.node import Node
from ..geometry.chain import Chain, EndPoints
from ..geometry.geometry_utils import (
    euclidean_distance_between_nodes,
    get_closest_dots_to_angle_on_radial_direction_sorted_by_ascending_distance_to_center,
)


def compute_interpolation_domain(
    endpoint: str, endpoint_cad1: Node, endpoint_cad2: Node, nr: int
) -> List[float]:
    """
    Compute the interpolation domain between two endpoints.

    Args:
        endpoint (str): endpoint type
        endpoint_cad1 (Node): first endpoint
        endpoint_cad2 (Node): second endpoint
        nr (int): number of radial distances

    Returns:
        List[float]: interpolation domain
    """
    interpolation_domain = []

    step = 360 / nr if endpoint == EndPoints.B else -360 / nr
    current_angle = endpoint_cad1.angle

    while current_angle % 360 != endpoint_cad2.angle:
        current_angle += step
        current_angle = current_angle % 360
        interpolation_domain.append(current_angle)

    return interpolation_domain[:-1]


def from_polar_to_cartesian(
    r: float, angulo: float, centro: Tuple[float, float]
) -> Tuple[float, float]:
    """
    Convert polar coordinates to cartesian coordinates.

    Args:
        r (float): radial distance
        angulo (float): angle in degrees
        centro (Tuple[float, float]): center coordinates

    Returns:
        Tuple[float, float]: cartesian coordinates
    """
    y = centro[0] + r * np.cos(angulo * np.pi / 180)
    x = centro[1] + r * np.sin(angulo * np.pi / 180)

    return (y, x)


def generate_node_list_between_two_support_chains_and_two_radial_distances(
    r2_ratio: float,
    r1_ratio: float,
    total_nodes: int,
    interpolation_angle_domain: List[float],
    center: Tuple[float, float],
    inward_chain: Chain,
    outward_chain: Chain,
    chain: Chain,
) -> List[Node]:
    """
    Generate a list of nodes between two support chains and two radial distances.

    Args:
        r2_ratio (float): radial distance of the last node
        r1_ratio (float): radial distance of the first node
        total_nodes (int): total nodes to generate
        interpolation_angle_domain (List[float]): radii angle of the nodes to generate
        center (Tuple[float, float]): center node of the disk
        inward_chain (Chain): inward support chain
        outward_chain (Chain): outward support chain
        chain (Chain): chain to be completed

    Returns:
        List[Node]: generated nodes list
    """
    cad_id = chain.id
    generated_node_list = []
    m = (r2_ratio - r1_ratio) / total_nodes
    n = r1_ratio

    for idx_current_node, angle in enumerate(interpolation_angle_domain):
        dot_list_in_radial_direction = get_closest_dots_to_angle_on_radial_direction_sorted_by_ascending_distance_to_center(
            [inward_chain], angle % 360
        )
        support_node = dot_list_in_radial_direction[0]
        radio_init = support_node.radial_distance
        dot_list_in_radial_direction = get_closest_dots_to_angle_on_radial_direction_sorted_by_ascending_distance_to_center(
            [outward_chain], angle % 360
        )
        support_node = dot_list_in_radial_direction[0]
        radio_superior = support_node.radial_distance
        radial_distance_between_chains = radio_superior - radio_init

        radio_inter = (
            m * (idx_current_node) + n
        ) * radial_distance_between_chains + radio_init
        i, j = from_polar_to_cartesian(radio_inter, angle % 360, center)
        i = i if i < chain.img_height else chain.img_height - 1
        j = j if j < chain.img_width else chain.img_width - 1

        # radios.append(radio_inter)
        params = {
            "x": j,
            "y": i,
            "angle": angle % 360,
            "radial_distance": radio_inter,
            "chain_id": cad_id,
        }

        node = Node(**params)
        generated_node_list.append(node)

    return generated_node_list


def generate_nodes_list_between_two_radial_distances(
    r2: float,
    r1: float,
    total_nodes: int,
    interpolation_angular_domain: List[float],
    center: Tuple[float, float],
    sign: int,
    ch_i: Optional[Chain],
    ch_j: Chain,
) -> List[Node]:
    """
    Generate a list of nodes between two radial distances.

    Args:
        r2 (float): radial distance of the last node
        r1 (float): radial distance of the first node
        total_nodes (int): total nodes to generate
        interpolation_angular_domain (List[float]): radii angle of the nodes to generate
        center (Tuple[float, float]): center node of the disk
        sign (int): sign of the radial distance
        ch_i (Optional[Chain]): inward support chain
        ch_j (Chain): outward support chain

    Returns:
        List[Node]: generated nodes list
    """
    cad_id = ch_j.id
    l_generated_node = []
    m = (r2 - r1) / (total_nodes - 0)
    n = r1 - m * 0
    for current_idx_node, angle in enumerate(interpolation_angular_domain):
        if ch_i is not None:
            dot_list_in_radial_direction = get_closest_dots_to_angle_on_radial_direction_sorted_by_ascending_distance_to_center(
                [ch_i], angle % 360
            )

            support_node = dot_list_in_radial_direction[0]
            radio_init = support_node.radial_distance
            # radio_init = r1
        else:
            radio_init = 0

        radio_inter = sign * (m * (current_idx_node) + n) + radio_init
        i, j = from_polar_to_cartesian(radio_inter, angle % 360, center)
        i = i if i < ch_j.img_height else ch_j.img_height - 1
        j = j if j < ch_j.img_width else ch_j.img_width - 1

        l_generated_node.append(
            Node(
                **{
                    "x": j,
                    "y": i,
                    "angle": angle % 360,
                    "radial_distance": radio_inter,
                    "chain_id": cad_id,
                }
            )
        )

    return l_generated_node


def get_radial_distance_to_chain(chain: Chain, dot: Node) -> float:
    """
    Get the radial distance to a chain.

    Args:
        chain (Chain): chain to compute the radial distance
        dot (Node): node to compute the radial distance

    Returns:
        float: radial distance
    """
    dot_list_in_radial_direction = get_closest_dots_to_angle_on_radial_direction_sorted_by_ascending_distance_to_center(
        [chain], dot.angle
    )
    soporte_pto1 = dot_list_in_radial_direction[0]
    rii = euclidean_distance_between_nodes(soporte_pto1, dot)

    return rii


def compute_radial_ratio(
    cadena_inferior: Chain, cadena_superior: Chain, dot: Node
) -> float:
    """
    Compute the radial ratio between two chains.

    Args:
        cadena_inferior (Chain): inferior chain
        cadena_superior (Chain): superior chain
        dot (Node): node to compute the radial distance

    Returns:
        float: radial ratio
    """
    r1_inferior = get_radial_distance_to_chain(cadena_inferior, dot)
    r1_superior = get_radial_distance_to_chain(cadena_superior, dot)

    return r1_inferior / (r1_superior + r1_inferior)


def interpolate_in_angular_domain_via_2_chains(
    inward_support_chain: Chain,
    outward_support_chain: Chain,
    ch1_endpoint: Node,
    ch2_endpoint: Node,
    endpoint: str,
    ch1: Chain,
    node_list: List[Node],
) -> None:
    """
    Interpolate in the angular domain using two chains.

    Args:
        inward_support_chain (Chain): inward support chain
        outward_support_chain (Chain): outward support chain
        ch1_endpoint (Node): chain 1 endpoint
        ch2_endpoint (Node): chain 2 endpoint
        endpoint (str): endpoint type
        ch1 (Chain): chain 1
        node_list (List[Node]): list of nodes

    Returns:
        None
    """
    # 1. Domain angle interpolation
    domain_angle_interpolation = compute_interpolation_domain(
        endpoint, ch1_endpoint, ch2_endpoint, ch1.nr
    )
    center = ch1.center

    # 2. Compute radial ratio
    r1_ratio = compute_radial_ratio(
        inward_support_chain, outward_support_chain, ch1_endpoint
    )
    r2_ratio = compute_radial_ratio(
        inward_support_chain, outward_support_chain, ch2_endpoint
    )

    # 3. Generate nodes
    total_nodes = len(domain_angle_interpolation)
    if total_nodes == 0:
        return

    generated_nodes = (
        generate_node_list_between_two_support_chains_and_two_radial_distances(
            r2_ratio,
            r1_ratio,
            total_nodes,
            domain_angle_interpolation,
            center,
            inward_support_chain,
            outward_support_chain,
            ch1,
        )
    )

    node_list += generated_nodes

    return


def domain_interpolation(
    ch_i: Optional[Chain],
    ch_j_endpoint: Node,
    ch_k_endpoint: Node,
    endpoint: int,
    ch_j: Chain,
    l_nodes: List[Node],
) -> None:
    """
    Interpolate between endpoint ch_j_endpoint and ch_k_endpoint using ch_i as support ch_j. Ch_j is the source ch_j to
    be connected

    Args:
        ch_i (Optional[Chain]): support ch_i
        ch_j_endpoint (Node): ch_j endpoint
        ch_k_endpoint (Node): ch_k endpoint
        endpoint (int): endpoint type
        ch_j (Chain): ch_j to be connected
        l_nodes (List[Node]): list of nodes

    Returns:
        None
    """
    domain_angles = compute_interpolation_domain(
        endpoint, ch_j_endpoint, ch_k_endpoint, ch_j.nr
    )
    center = ch_j.center

    if ch_i is not None:
        dot_list_in_radial_direction = get_closest_dots_to_angle_on_radial_direction_sorted_by_ascending_distance_to_center(
            [ch_i], ch_j_endpoint.angle
        )
        node1_support = dot_list_in_radial_direction[0]
        r1 = euclidean_distance_between_nodes(node1_support, ch_j_endpoint)
        dot_list_in_radial_direction = get_closest_dots_to_angle_on_radial_direction_sorted_by_ascending_distance_to_center(
            [ch_i], ch_k_endpoint.angle
        )
        node2_support = dot_list_in_radial_direction[0]
        sign = (
            -1 if node2_support.radial_distance > ch_k_endpoint.radial_distance else +1
        )
        r2 = euclidean_distance_between_nodes(node2_support, ch_k_endpoint)
    else:
        r1 = ch_j_endpoint.radial_distance
        r2 = ch_k_endpoint.radial_distance
        sign = 1

    total_nodes = len(domain_angles)
    if total_nodes == 0:
        return

    l_generated_nodes = generate_nodes_list_between_two_radial_distances(
        r2, r1, total_nodes, domain_angles, center, sign, ch_i, ch_j
    )

    l_nodes += l_generated_nodes

    return


def complete_chain_using_2_support_ring(
    inward_chain: Chain, outward_chain: Chain, chain: Chain
) -> bool:
    """
    Complete ch_i using two support rings

    Args:
        inward_chain (Chain): inward chain
        outward_chain (Chain): outward chain
        chain (Chain): chain to be completed

    Returns:
        bool: boolean value indicating if the border has changed
    """
    ch1_endpoint = chain.extB
    ch2_endpoint = chain.extA
    endpoint = EndPoints.B
    generated_nodes = []

    interpolate_in_angular_domain_via_2_chains(
        inward_chain,
        outward_chain,
        ch1_endpoint,
        ch2_endpoint,
        endpoint,
        chain,
        generated_nodes,
    )

    change_border = chain.add_nodes_list(generated_nodes)

    return change_border


def complete_chain_using_support_ring(support_chain: Chain, ch1: Chain) -> bool:
    """
    Complete ch_i using a support ring

    Args:
        support_chain (Chain): support chain
        ch1 (Chain): chain to be completed

    Returns:
        bool: boolean value indicating if the border has changed
    """
    ch1_endpoint = ch1.extB
    ch2_endpoint = ch1.extA
    endpoint = EndPoints.B
    generated_list_nodes = []
    domain_interpolation(
        support_chain, ch1_endpoint, ch2_endpoint, endpoint, ch1, generated_list_nodes
    )
    change_border = ch1.add_nodes_list(generated_list_nodes)

    return change_border


def connect_2_chain_via_inward_and_outward_ring(
    outward_chain: Chain,
    inward_chain: Chain,
    chain1: Chain,
    chain2: Chain,
    node_list: List[Node],
    endpoint: int,
) -> Tuple[List[Node], bool]:
    """
    Connect 2 ch_i via inward and outward ring

    Args:
        outward_chain (Chain): outward chain
        inward_chain (Chain): inward chain
        chain1 (Chain): chain 1
        chain2 (Chain): chain 2
        node_list (List[Node]): list of nodes
        endpoint (int): endpoint type

    Returns:
        Tuple[List[Node], bool]: generated nodes list and boolean value indicating if the border has changed
    """
    ch1_endpoint = chain1.extA if endpoint == EndPoints.A else chain1.extB
    ch2_endpoint = chain2.extB if endpoint == EndPoints.A else chain2.extA

    # 1.0
    generated_node_list = []
    interpolate_in_angular_domain_via_2_chains(
        inward_chain,
        outward_chain,
        ch1_endpoint,
        ch2_endpoint,
        endpoint,
        chain1,
        generated_node_list,
    )
    node_list += generated_node_list

    # 2.0
    chain_2_nodes = []
    chain_2_nodes += chain2.l_nodes
    for node in chain_2_nodes:
        node.chain_id = chain1.id

    change_border = chain1.add_nodes_list(chain_2_nodes + generated_node_list)

    return generated_node_list, change_border



import json
import shutil
from pathlib import Path
from typing import Dict, Any
import cv2
import numpy as np


def clear_directory(dir_path: Path) -> None:
    """
    Clear all contents of a directory.

    Args:
        dir_path (Path): Directory to clear
    """
    if not dir_path.exists():
        return

    for item in dir_path.iterdir():
        if item.is_file():
            item.unlink()
        elif item.is_dir():
            shutil.rmtree(item)


def load_json(filepath: Path) -> dict:
    """
    Load a JSON file.

    Args:
        filepath (Path): Path to JSON file

    Returns:
        dict: Loaded JSON data
    """
    with filepath.open("r") as f:
        return json.load(f)


def write_json(data: Dict[str, Any], filepath: Path) -> None:
    """
    Write data to a JSON file.

    Args:
        data (Dict[str, Any]): Data to write
        filepath (Path): Output file path
    """
    with open(filepath, "w") as f:
        json.dump(data, f, indent=4)


def load_image(filepath: Path) -> cv2.typing.MatLike:
    """
    Load an image file.

    Args:
        filepath (Path): Path to image file

    Returns:
        cv2.typing.MatLike: Image as RGB numpy array
    """
    img = cv2.imread(str(filepath))
    return cv2.cvtColor(img, cv2.COLOR_BGR2RGB)


def ensure_directory(dir_path: Path, clear: bool = False) -> Path:
    """
    Ensure a directory exists, optionally clearing it first.

    Args:
        dir_path (Path): Directory path
        clear (bool): Whether to clear existing contents

    Returns:
        Path: Resolved directory path
    """
    dir_path = dir_path.resolve()

    if dir_path.exists() and clear:
        clear_directory(dir_path)

    dir_path.mkdir(parents=True, exist_ok=True)
    return dir_path
